#!/usr/bin/perl -w
use strict;
use List::Util qw(first max maxstr min minstr reduce shuffle sum);

# Please install this package to generate random numbers
# rand() function does not generate good randon numbers
use Math::Random::Secure qw(rand);
srand( time^( $$ + ( $$ << 15 ) ) );

my $infile_study = $ARGV[0]; # The list of the genes used in the study / or "null" for all genes
                             #    all gene IDs must be Entrez Gene IDs (all as positive integers)
my $infile_mirts = $ARGV[1]; # The miRNA family - target gene table (../_data/miRNA/TargetScan/v6.2)
my $n_permute = $ARGV[2]; # The runs of permutation
                          #   -1 for generating the original gene set file
                          #   >0 (N = 1000 or 10000) for generateing N random gene sets as FAME algorithm
my $min_target_number = 20; # If the number of genes is less than this number, skip it

my %study1 = (); # the hash for the studied gene list
# If $infile_study equals "null", all genes are included
#    or include the genes in the input gene list file
if( lc( $infile_study ) ne "null" ) {
	open( in1, "<$infile_study" );
	while( my $line1 = <in1> ) {
		chomp( $line1 );
		my ( $id1, $desc1 ) = split( /\t/, $line1, 2 );
		
		# if ID contains non-digital char, skip
		if( $id1 =~ m/\D/ ) {
			next;
		}
		
		$study1{ $id1 } = "yes";
	}
	close( in1 );
}

my %mirfam_target_score = (); # hash key as "miRFam<\tab>EntrezGeneID"; value as "TargetScan summed context score"
my %mirfam_target_number = (); # hash key as "miRFam"; value as "the number of targets"
open( in2, "<$infile_mirts" );
my $header2 = <in2>;
while( my $line2 = <in2> ) {
	chomp( $line2 );
	my ( $mirfam1, $id1, $symbol1, $score1, $repmir1 ) = split( /\t/, $line2 );
	
	if( lc( $infile_study ) ne "null" && !exists( $study1{ $id1 } ) ) {
		next;
	}
	
	$mirfam_target_score{ $mirfam1."\t".$id1 } = $score1;
	if( !exists( $mirfam_target_number{ $mirfam1 } ) ) {
		$mirfam_target_number{ $mirfam1 } = 1;
	}
	else {
		$mirfam_target_number{ $mirfam1 }++;
	}
}
close( in2 );

# miRNA:Target with larger negative values (higher confidence) should rank at top
my @sort_key = sort { $mirfam_target_score{ $a } <=> $mirfam_target_score{ $b } } keys %mirfam_target_score;

# delete the miRNA with less than $min_target_number target genes
my %mirfam_delete = ();
my @mirfam_arr = sort keys %mirfam_target_number;
for( my $i = 0; $i <= $#mirfam_arr; $i++ ) {
	my $key = $mirfam_arr[$i];
	if( $mirfam_target_number{ $key } < $min_target_number ) {
		$mirfam_delete{ $key } = "yes";
		#print $key, "\n";
	}
}

for( my $i = 0; $i <= $#sort_key; $i++ ) {
	my ( $mirfam1, $target1 ) = split( /\t/, $sort_key[$i] );
	if( exists( $mirfam_delete{ $mirfam1 } ) ) {
		delete( $mirfam_target_score{ $sort_key[$i] } );
		splice( @sort_key, $i, 1 );
		$i--;
	}
}

### ---------- Calculate the normalized scores for all miRFam:Target pairs ----------###
my $level1 = 5; # a in FAME: dicretize scores to "a" levels (bigger means higher confidence)
my $balance1 = 3; # b in FAME: the delta strength between two adjacent levels
my $n_edge = $#sort_key + 1;

my %level2edge = (); # hash key as the "score level"; value as the corresponding miRFam:Target pairs seperated by ","
my %edge2level = (); # hash key as the "miRFam<\tab>Target" pair; value as the "score level" (dicretized TargetScan scores)

# Calculate from the top of the sorted miRFam:Target list
for( my $i = 0; $i <= $#sort_key; $i++ ) {
	my $key = $sort_key[$i];
	my $rank_percent = 1 - ( $i + 1 ) / $n_edge; # The top pairs hould have larger scores, so use ( 1 - $top_percentage )
	
	# FAME normalized score: score = 1 + int( r% * a ) * b
	my $edge_score = 1 + int( $rank_percent * $level1 ) * $balance1;
	
	if( !exists( $level2edge{ $edge_score } ) ) {
		$level2edge{ $edge_score } = $key;
	}
	else {
		$level2edge{ $edge_score } = $level2edge{ $edge_score }.",".$key;
		###if( $key eq "miR-1249\t9145" ) { print "$edge_score\n"; }
	}
	
	$edge2level{ $key } = $edge_score;
}

### print $level2edge{ "10" }, "\n";

# if $n_permute <= 0, generate the original gene set file
if( $n_permute <= 0 ) {
	my %mirfam_set = (); # hash key as miRFam; value as "Target:score" list seperated by <\tab>
	
	# map miRFam:Target hash as miRFam hash
	table2geneset( \%edge2level, \%mirfam_set );
	
	my @key2 = sort keys %mirfam_set;
	for( my $i = 0; $i <= $#key2; $i++ ) {
		my $key = $key2[$i];
		my @target1 = split( /\t/, $mirfam_set{ $key } );
		my $n_target1 = $#target1 + 1;
		
		# ERROR checking for the size of miRNA target gene set
		if( $n_target1 != $mirfam_target_number{ $key } ) {
			print "ERROR for checking the size of gene set: $key";
			exit;
		}
				
		print join( "\t", $key, $n_target1, $mirfam_set{ $key } ), "\n";
	}
}
else {
	for( my $i_run = 1; $i_run <= $n_permute; $i_run++ ) {
		print ">Permuting RUN $i_run.\n";
		
		my %r_edge2level = %edge2level; # The permuted "miRFam<\tab>Target" - "score level" hash table
		
		# Permute the edges with equal scores (check for equal genes of the same miRFam)
		for( my $i_level = 0; $i_level < $level1; $i_level++ ) {
			my $select_score = 1 + $i_level * $balance1; # current score level ready for permuting
			my @edge = split( /,/, $level2edge{ $select_score } ); # all edges for the current score level
			my $i_permute_level = 5 * ( $#edge + 1 ); # the times for permuting
			
			while( $i_permute_level > 0 ) {
				my $select_edge1 = int( rand( $#edge + 1 ) ); # the first selected miRFam:target edge
				my $select_edge2 = int( rand( $#edge + 1 ) ); # the second edge
			
				while( $select_edge1 == $select_edge2 ) { # the two edges should be different
					$select_edge2 = int( rand( $#edge + 1 ) );
				}
			  
				my ( $mir1, $id1 ) = split( /\t/, $edge[ $select_edge1 ] );
				my ( $mir2, $id2 ) = split( /\t/, $edge[ $select_edge2 ] );
				
				# checking for equal genes of the same miRFam
				#   miRFam1 should not include ID2 as target - miRFam2 should not include ID1 as target
				if( exists( $r_edge2level{ "$mir1\t$id2" } ) || exists( $r_edge2level{ "$mir2\t$id1" } ) ) {
					next;
				}
			  
			  # Exchange the target genes of the two selected edges
				delete( $r_edge2level{ $edge[ $select_edge1 ] } );
				delete( $r_edge2level{ $edge[ $select_edge2 ] } );
			  
				$edge[ $select_edge1 ] = "$mir1\t$id2";
				$edge[ $select_edge2 ] = "$mir2\t$id1";
			  
				$r_edge2level{ $edge[ $select_edge1 ] } = $select_score;
				$r_edge2level{ $edge[ $select_edge2 ] } = $select_score;
			
				$i_permute_level--;
			}
			
			#$r_level2edge{ $select_score } = join( ",", @edge );
		}
		
		my %r_mirfam_set = ();
		table2geneset( \%r_edge2level, \%r_mirfam_set );
	
		my @key2 = sort keys %r_mirfam_set;
		for( my $i = 0; $i <= $#key2; $i++ ) {
			my $key = $key2[$i];
			my @target1 = split( /\t/, $r_mirfam_set{ $key } );
			my $n_target1 = $#target1 + 1;
			
			if( $n_target1 != $mirfam_target_number{ $key } ) {
				print "ERROR for checking the size of gene set: $key";
				exit;
			}
			
			print join( "\t", $key, $n_target1, $r_mirfam_set{ $key } ), "\n";
		}
		
		print "//\n";
	}
}

sub table2geneset {
	my ( $hash_edge, $hash_mirfam ) = @_;
	my @edge1 = sort keys %$hash_edge;
	
	for( my $i = 0; $i <= $#edge1; $i++ ) {
		my ( $mir1, $id1 ) = split( /\t/, $edge1[$i] );
		
		if( ! exists( $hash_mirfam->{ $mir1 } ) ) {
			$hash_mirfam->{ $mir1 } = $id1.":".$hash_edge->{ $edge1[$i] };
		}
		else {
			$hash_mirfam->{ $mir1 } = $hash_mirfam->{ $mir1 }."\t".$id1.":".$hash_edge->{ $edge1[$i] };
		}
	}
}

